'================================================
' This class is based on the SF SqlDataSourceItem class
' Created by Edhy Rijo on 4/28/2010
' Basic changes done:
'  Replaced all references from SqlCommand with MySqlCommand
'  Replaced all references of [ ] with ` in the Build... commands
'  In the BuildInsertCommand_NonSP method replaced the statement to get the PK for the inserted
'  record using MySql function "SELECT LAST_INSERT_ID() "
' This class was created using MySqlConnector version 6.3 at http://dev.mysql.com/downloads/connector/

Imports System.Text
Imports System.Data.Common
Imports MicroFour.StrataFrame.Data
Imports System.Collections.Generic
Imports System.Data
Imports MySql.Data
Imports MySql.Data.MySqlClient



Namespace Data
    Public Class MySqlDataSourceItem
        Inherits DbDataSourceItem

#Region " Constructors "

        ''' <summary>
        ''' Initializes a new instance of the MySqlDataSourceItem class.
        ''' </summary>
        ''' <param name="DataSourceKey"></param>
        ''' <remarks></remarks>
        Public Sub New(ByVal DataSourceKey As String)
            MyBase.New(DataSourceKey, "")
        End Sub

        ''' <summary>
        ''' Initializes a new instance of the MySqlDataSourceItem class.
        ''' </summary>
        ''' <param name="DataSourceKey"></param>
        ''' <param name="ConnectionString"></param>
        ''' <remarks></remarks>
        Public Sub New(ByVal DataSourceKey As String, ByVal ConnectionString As String)
            MyBase.New(DataSourceKey, ConnectionString)
        End Sub

#End Region

#Region " Private Fields "

        Private _CachedUpdateCommands As New Dictionary(Of DataTable, MySqlCommand)
        Private _CachedDeleteCommands As New Dictionary(Of DataTable, MySqlCommand)
        Private _CachedInsertCommands As New Dictionary(Of DataTable, MySqlCommand)

#End Region

#Region " Public Properties "

        ''' <summary>
        ''' Gets a value that indicates that this data source item supports retrieval of blob fields 
        ''' in pieces.
        ''' </summary>
        ''' <value></value>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Public Overrides ReadOnly Property SupportsIncrementalBlobRetrieval() As Boolean
            Get
                Return True
            End Get
        End Property

#End Region

#Region " Private Methods "

        Private Function BuildDeleteCommand(ByVal QueryInfo As QueryInformation) As MySqlCommand
            '-- Determine whether this is a stored procedure or parameterized command
            If QueryInfo.UseStoredProcedure Then
                Return Me.BuildDeleteCommand_SP(QueryInfo)
            Else
                Return Me.BuildDeleteCommand_NonSP(QueryInfo)
            End If
        End Function

        Private Function BuildDeleteCommand_NonSP(ByVal QueryInfo As QueryInformation) As MySqlCommand
            '-- Establish locals
            Dim loReturn As New MySqlCommand
            Dim loParam As MySqlParameter
            Dim lcField As String
            Dim loQuery As New StringBuilder(1024)
            Dim loType As MySqlDbType

            '-- Begin building the command
            loQuery.Append("DELETE FROM `")
            loQuery.Append(QueryInfo.TableSchema)
            loQuery.Append("`.`")
            loQuery.Append(QueryInfo.TableName)
            loQuery.Append("` WHERE ")

            '-- Add the primary key specification
            For Each lcField In QueryInfo.PrimaryKeyFields
                '-- Add the field
                loQuery.Append("`")
                loQuery.Append(lcField)
                loQuery.Append("` = @")
                loQuery.Append(lcField)
                loQuery.Append(" AND ")

                '-- Add the primary key parameter
                loParam = Me.CreateDbParameter(QueryInfo, lcField)
                loParam.ParameterName = "@" & lcField
                loParam.SourceColumn = lcField
                loParam.Direction = ParameterDirection.Input
                loReturn.Parameters.Add(loParam)
            Next

            '-- Delete the last "AND"
            loQuery.Length -= 5

            '-- Add the row version fields to the return value
            If QueryInfo.ConcurrencyType = ConcurrencyTypeOptions.OptimisticRowVersion OrElse _
                    QueryInfo.ConcurrencyType = ConcurrencyTypeOptions.OptimisticTimestamp Then

                For Each lcField In QueryInfo.RowVersionFields
                    '-- Only check the NULL on the parameter
                    loQuery.Append(" AND ((`")
                    loQuery.Append(lcField)
                    loQuery.Append("` = @")
                    loQuery.Append(lcField)
                    loQuery.Append(") OR (@")
                    loQuery.Append(lcField)
                    loQuery.Append(" IS NULL))")

                    '-- Append the parameter
                    loParam = Me.CreateDbParameter(QueryInfo, lcField)
                    loParam.ParameterName = "@" & lcField
                    loParam.SourceColumn = lcField
                    loParam.Direction = ParameterDirection.Input
                    loReturn.Parameters.Add(loParam)
                Next
            ElseIf QueryInfo.ConcurrencyType = ConcurrencyTypeOptions.OptimisticAllFields Then
                For Each lcField In QueryInfo.Fields
                    '-- Add the field to the query
                    '-- Make sure the field is not an image field
                    'Try
                    '    loType = CType(QueryInfo.FieldNativeDbTypes(lcField), MySqlDbType)
                    '    If loType = MySqlDbType.Image Then
                    '        Continue For
                    '    End If
                    'Catch ex As Exception
                    '    loType = MySqlDbType.VarChar
                    'End Try

                    loQuery.Append(" AND ((`")
                    loQuery.Append(lcField)
                    If (loType = MySqlDbType.Text) Then
                        loQuery.Append("` LIKE @")
                    Else
                        loQuery.Append("` = @")
                    End If
                    loQuery.Append(lcField)
                    loQuery.Append("_ORG) OR ((@")
                    loQuery.Append(lcField)
                    loQuery.Append("_ORG IS NULL) AND (`")
                    loQuery.Append(lcField)
                    loQuery.Append("` IS NULL)))")

                    '-- Add the original field value parameter
                    loParam = Me.CreateDbParameter(QueryInfo, lcField)
                    loParam.ParameterName = "@" & lcField & "_ORG"
                    loParam.SourceColumn = lcField
                    loParam.SourceVersion = DataRowVersion.Original
                    loParam.Direction = ParameterDirection.Input
                    loReturn.Parameters.Add(loParam)
                Next
            End If

            '-- Finish off the command
            loQuery.AppendLine(";")

            '-- Set the text on the command
            loReturn.CommandText = ReplaceTokens(loQuery.ToString(), QueryInfo)
            loReturn.CommandType = CommandType.Text

            '-- Return the created command
            Return loReturn
        End Function

        Private Function BuildDeleteCommand_SP(ByVal QueryInfo As QueryInformation) As MySqlCommand
            '-- Establish locals
            Dim loReturn As New MySqlCommand()
            Dim loParam As MySqlParameter
            Dim lcField As String
            Dim lnCnt As Integer

            '-- Set the stored procedure name
            loReturn.CommandText = QueryInfo.StoredProcedureName
            loReturn.CommandType = CommandType.StoredProcedure

            '-- Add the primary key parameters
            For lnCnt = 0 To QueryInfo.PrimaryKeyFields.Length - 1
                lcField = QueryInfo.PrimaryKeyFields(lnCnt)
                loParam = Me.CreateDbParameter(QueryInfo, lcField)
                loParam.ParameterName = "@" & lcField
                loParam.SourceColumn = lcField
                loParam.Direction = ParameterDirection.Input
                loReturn.Parameters.Add(loParam)
            Next

            '-- Check the concurrency type
            '-- Add the row version fields
            If QueryInfo.ConcurrencyType = ConcurrencyTypeOptions.OptimisticRowVersion OrElse _
                QueryInfo.ConcurrencyType = ConcurrencyTypeOptions.OptimisticTimestamp Then

                '-- Cycle through the fields
                For Each lcField In QueryInfo.RowVersionFields
                    '-- Add the field parameter
                    loParam = Me.CreateDbParameter(QueryInfo, lcField)
                    loParam.ParameterName = "@" & lcField
                    loParam.SourceColumn = lcField
                    loParam.Direction = ParameterDirection.Input
                    loReturn.Parameters.Add(loParam)
                Next
            ElseIf QueryInfo.ConcurrencyType = ConcurrencyTypeOptions.OptimisticAllFields Then
                '-- Add all of the fields with the _ORG tag
                For Each lcField In QueryInfo.Fields
                    '-- Add the field parameter
                    loParam = Me.CreateDbParameter(QueryInfo, lcField)
                    loParam.ParameterName = "@" & lcField & "_ORG"
                    loParam.SourceColumn = lcField
                    loParam.SourceVersion = DataRowVersion.Original
                    loParam.Direction = ParameterDirection.Input
                    loReturn.Parameters.Add(loParam)
                Next
            End If

            '-- Return the created command
            Return loReturn
        End Function

        Private Function BuildInsertCommand(ByVal QueryInfo As QueryInformation) As MySqlCommand
            '-- Determine whether this is a stored procedure or parameterized command
            If QueryInfo.UseStoredProcedure Then
                Return Me.BuildInsertCommand_SP(QueryInfo)
            Else
                Return Me.BuildInsertCommand_NonSP(QueryInfo)
            End If
        End Function

        Private Function BuildInsertCommand_NonSP(ByVal QueryInfo As QueryInformation) As MySqlCommand
            '-- Establish locals
            Dim loReturn As New MySqlCommand()
            Dim loParam As MySqlParameter
            Dim lcField As String
            Dim loQuery As New StringBuilder(1024)

            '-- Begin by building the command
            loQuery.Append("INSERT INTO `")
            loQuery.Append(QueryInfo.TableSchema)
            loQuery.Append("`.`")
            loQuery.Append(QueryInfo.TableName)
            loQuery.Append("` (")

            '-- Add the primary key field if necessary
            If Not QueryInfo.PrimaryKeyFieldIsAutoInc Then
                For Each lcField In QueryInfo.PrimaryKeyFields
                    loQuery.Append("`")
                    loQuery.Append(lcField)
                    loQuery.Append("`, ")
                Next
            End If

            '-- Add the field names
            For Each lcField In QueryInfo.Fields
                loQuery.Append("`")
                loQuery.Append(lcField)
                loQuery.Append("`, ")
            Next

            '-- Add the row version fields if we need to
            If (QueryInfo.ConcurrencyType = ConcurrencyTypeOptions.OptimisticRowVersion) Then
                '-- Add the field
                For Each lcField In QueryInfo.RowVersionFields
                    loQuery.Append("`")
                    loQuery.Append(lcField)
                    loQuery.Append("`, ")
                Next
            End If

            '-- Remove the last 2 characters ", "
            loQuery.Length -= 2

            '-- Close off the parenthasis and begin the VALUES portion of the command
            loQuery.Append(") VALUES (")

            '-- Add the primary key field parameter if necessary
            If Not QueryInfo.PrimaryKeyFieldIsAutoInc Then
                For Each lcField In QueryInfo.PrimaryKeyFields
                    '-- Add the field to the query
                    loQuery.Append("@")
                    loQuery.Append(lcField)
                    loQuery.Append(", ")

                    '-- Add the parameter to the command
                    loParam = Me.CreateDbParameter(QueryInfo, lcField)
                    loParam.ParameterName = "@" & lcField
                    loParam.SourceColumn = lcField
                    loParam.Direction = ParameterDirection.Input
                    loReturn.Parameters.Add(loParam)
                Next
            End If

            '-- Add the field parameters 
            For Each lcField In QueryInfo.Fields
                '-- Add the field to the query
                loQuery.Append("@")
                loQuery.Append(lcField)
                loQuery.Append(", ")

                '-- Add the parameter to the command
                loParam = Me.CreateDbParameter(QueryInfo, lcField)
                loParam.ParameterName = "@" & lcField
                loParam.SourceColumn = lcField
                loParam.Direction = ParameterDirection.Input
                loReturn.Parameters.Add(loParam)
            Next

            '-- Add the row version fields if we need to
            If (QueryInfo.ConcurrencyType = ConcurrencyTypeOptions.OptimisticRowVersion) Then
                '-- Add the field
                For Each lcField In QueryInfo.RowVersionFields
                    '-- Add the field to the query
                    loQuery.Append("@")
                    loQuery.Append(lcField)
                    loQuery.Append(", ")

                    '-- Add the parameter to the command
                    loParam = Me.CreateDbParameter(QueryInfo, lcField)
                    loParam.ParameterName = "@" & lcField
                    loParam.SourceColumn = lcField
                    loParam.Direction = ParameterDirection.InputOutput
                    loReturn.Parameters.Add(loParam)
                Next
            End If

            '-- Remove the last 2 characters ", "
            loQuery.Length -= 2

            '-- Finish off the command
            loQuery.AppendLine(");")

            '-- Determine if we need to re-select anything from the newly inserted record
            If (QueryInfo.RowVersionFields.Count > 0) OrElse _
                (QueryInfo.PrimaryKeyFieldIsAutoInc) Then

                '-- Start the select command
                '   Edhy Rijo 4/29/2010 this is all need to get the last inserted record in MySQL
                loQuery.Append("SELECT LAST_INSERT_ID() ")

                'loQuery.Append("SELECT ")

                '-- See if we need to add the identity return
                If QueryInfo.PrimaryKeyFieldIsAutoInc Then
                    '-- Make sure there is only 1 primary key field
                    If QueryInfo.PrimaryKeyFields.Length <> 1 Then
                        Throw New Exception("A business object with PrimaryKeyIsAutoIncremented = True must have exactly 1 primary key field.")
                    End If

                    '-- Edhy Rijo since LAST_INSERT_ID() returns a field name with "LAST_INSERT_ID"
                    '   we need to add the AS clause to assign the PK field name instead.
                    '-- Append the identity retrieval
                    loQuery.Append(" AS `")
                    loQuery.Append(QueryInfo.PrimaryKeyFields(0))
                    loQuery.Append("` ")
                End If

                ''-- See if we need to append the row version retrieval
                'If QueryInfo.RowVersionFields.Count > 0 Then
                '    '-- Append each field to retrieve
                '    For Each lcField In QueryInfo.RowVersionFields
                '        '-- Append the field
                '        loQuery.Append("`")
                '        loQuery.Append(lcField)
                '        loQuery.Append("`, ")
                '    Next
                'End If

                ''-- Cut off the last 2 characters ", "
                'loQuery.Length -= 2

                ''-- Append the tablename and pk value
                'loQuery.Append(" FROM `")
                'loQuery.Append(QueryInfo.TableSchema)
                'loQuery.Append("`.`")
                'loQuery.Append(QueryInfo.TableName)
                'loQuery.Append("` WHERE ")
                'loQuery.Append(QueryInfo.PrimaryKeyFields(0))
                'loQuery.Append(" = ")
                'If QueryInfo.PrimaryKeyFieldIsAutoInc Then
                '    loQuery.Append("SCOPE_IDENTITY()")
                'Else
                '    loQuery.Append("@")
                '    loQuery.Append(QueryInfo.PrimaryKeyFields(0))
                'End If

                '-- Close off the select command
                loQuery.AppendLine(";")
            End If

            '-- Build the command
            loReturn.CommandText = ReplaceTokens(loQuery.ToString(), QueryInfo)
            loReturn.CommandType = CommandType.Text

            '-- Return the value
            Return loReturn
        End Function

        Private Function BuildInsertCommand_SP(ByVal QueryInfo As QueryInformation) As MySqlCommand
            '-- Establish locals
            Dim loReturn As New MySqlCommand()
            Dim loParam As MySqlParameter
            Dim lcField As String

            '-- Set the command type and text
            loReturn.CommandType = CommandType.StoredProcedure
            loReturn.CommandText = QueryInfo.StoredProcedureName

            '-- Add the parameters
            For Each lcField In QueryInfo.Fields
                '-- Add the parameter
                loParam = Me.CreateDbParameter(QueryInfo, lcField)
                loParam.ParameterName = "@" & lcField
                loParam.SourceColumn = lcField
                loParam.Direction = ParameterDirection.Input
                loReturn.Parameters.Add(loParam)
            Next

            '-- Add the parameter for the primary key
            For Each lcField In QueryInfo.PrimaryKeyFields
                loParam = Me.CreateDbParameter(QueryInfo, lcField)
                loParam.ParameterName = "@" & lcField
                loParam.SourceColumn = lcField
                If QueryInfo.PrimaryKeyFieldIsAutoInc Then
                    loParam.Direction = ParameterDirection.Output
                Else
                    loParam.Direction = ParameterDirection.Input
                End If
                loReturn.Parameters.Add(loParam)
            Next

            '-- Determine if we need to add the output parameters for the version
            For Each lcField In QueryInfo.RowVersionFields
                '-- Add the parameter
                loParam = Me.CreateDbParameter(QueryInfo, lcField)
                loParam.ParameterName = "@" & lcField
                loParam.SourceColumn = lcField
                loParam.Direction = ParameterDirection.Output
                loReturn.Parameters.Add(loParam)
            Next

            '-- Return the value
            Return loReturn
        End Function

        Private Function BuildSelectCommand(ByVal QueryInfo As QueryInformation) As MySqlCommand
            '-- Establish locals
            Dim loReturn As New MySqlCommand()
            Dim loParam As MySqlParameter
            Dim lcField As String
            Dim loQuery As New StringBuilder(1024)
            Dim loWhere As WhereStatement
            Dim lnParamCnt As Integer = 0
            Dim loValue As Object
            Dim lnCnt As Integer
            Dim lnPkCnt As Integer
            Dim lnValueCnt As Integer = 0

            '-- Begin building the command
            loQuery.Append("SELECT ")

            '-- Determine if we are DISTINCT
            If QueryInfo.IsDistinct Then
                loQuery.Append("DISTINCT ")
            End If

            '-- Determine if we need to add the TOP to the command
            If QueryInfo.QueryType = QueryTypeOptions.SELECT_TOP Then
                loQuery.Append("TOP ")
                loQuery.Append(QueryInfo.TopRowCount.ToString())
                loQuery.Append(" ")
            End If

            '-- Cycle through the fields and add them to the query
            If QueryInfo.Fields.Count > 0 Then
                For Each lcField In QueryInfo.Fields
                    loQuery.Append("`")
                    loQuery.Append(lcField)
                    loQuery.Append("`, ")
                Next

                '-- Remove the last two characters ", "
                loQuery.Length -= 2
            Else
                loQuery.Append("*")
            End If

            '-- Continue with the query
            loQuery.Append(" FROM `")
            loQuery.Append(QueryInfo.TableSchema)
            loQuery.Append("`.`")
            loQuery.Append(QueryInfo.TableName)
            loQuery.Append("` ")

            '-- Add the where clause
            If QueryInfo.WhereStatements.Count > 0 Then
                loQuery.Append("WHERE ")
            End If

            '-- Add the where statements
            For Each loWhere In QueryInfo.WhereStatements
                '-- Parse the compound primary key differently
                If loWhere.StatementType = WhereStatementTypeOptions.CompoundPrimaryKey Then
                    '-- Add the first parenthasis
                    loQuery.Append("(")

                    '-- Cycle through the primary key fields
                    For lnCnt = 0 To CInt(loWhere.Values.Count / loWhere.Fields.Length) - 1
                        '-- Add the opening parenthasis
                        loQuery.Append("(")

                        '-- Add the primary key specification
                        For lnPkCnt = 0 To loWhere.Fields.Length - 1
                            '-- Add the AND if necessary
                            If lnPkCnt > 0 Then
                                loQuery.Append(" AND ")
                            End If

                            '-- Get the field name
                            lcField = loWhere.Fields(lnPkCnt)

                            '-- Add the query
                            loQuery.Append(lcField)
                            loQuery.Append(" = @PARAM")
                            loQuery.Append(lnParamCnt.ToString())

                            '-- Add the parameter to the return command
                            loParam = Me.CreateDbParameter(QueryInfo, lcField)
                            loParam.ParameterName = "@PARAM" & lnParamCnt.ToString()
                            loParam.Value = loWhere.Values(lnValueCnt)
                            loParam.Direction = ParameterDirection.Input
                            loReturn.Parameters.Add(loParam)

                            '-- Increment the parameter counter and value counter
                            lnParamCnt += 1
                            lnValueCnt += 1
                        Next

                        '-- Add the closing parenthasis
                        loQuery.Append(") OR ")
                    Next

                    '-- Remove the last " OR "
                    loQuery.Length -= 4

                    '-- Append the last parenthasis
                    loQuery.Append(")")
                Else
                    '-- Add the field spec
                    loQuery.Append("`")
                    loQuery.Append(loWhere.Fields(0))
                    loQuery.Append("` ")

                    '-- Add the operator and values
                    Select Case loWhere.StatementType
                        Case WhereStatementTypeOptions.BeginsWith, WhereStatementTypeOptions.Contains, _
                                WhereStatementTypeOptions.EndsWith, WhereStatementTypeOptions.Like
                            loQuery.Append("LIKE ")

                        Case WhereStatementTypeOptions.Equals
                            loQuery.Append("= ")

                        Case WhereStatementTypeOptions.GreaterThan
                            loQuery.Append("> ")

                        Case WhereStatementTypeOptions.GreaterThanOrEqual
                            loQuery.Append(">= ")

                        Case WhereStatementTypeOptions.In
                            loQuery.Append("IN ")

                        Case WhereStatementTypeOptions.LessThan
                            loQuery.Append("< ")

                        Case WhereStatementTypeOptions.LessThanOrEqual
                            loQuery.Append("<= ")

                    End Select

                    '-- Append the values and parameters
                    Select Case loWhere.StatementType
                        Case WhereStatementTypeOptions.BeginsWith
                            '-- Append the parametername
                            loQuery.Append("@PARAM")
                            loQuery.Append(lnParamCnt.ToString())

                            '-- Add the parameter to the return command
                            loParam = Me.CreateDbParameter(QueryInfo, loWhere.Fields(0))
                            loParam.ParameterName = "@PARAM" & lnParamCnt.ToString()
                            loParam.Value = CType(loWhere.Values(0), String) & "%"
                            loParam.Direction = ParameterDirection.Input
                            '-- Remove the length if this is a fixed length string
                            Me.RemoveLength(loParam)
                            loReturn.Parameters.Add(loParam)

                            '-- Increment the parameter counter
                            lnParamCnt += 1

                        Case WhereStatementTypeOptions.Contains
                            '-- Append the parametername
                            loQuery.Append("@PARAM")
                            loQuery.Append(lnParamCnt.ToString())

                            '-- Add the parameter to the return command
                            loParam = Me.CreateDbParameter(QueryInfo, loWhere.Fields(0))
                            loParam.ParameterName = "@PARAM" & lnParamCnt.ToString()
                            loParam.Value = "%" & CType(loWhere.Values(0), String) & "%"
                            loParam.Direction = ParameterDirection.Input
                            '-- Remove the length if this is a fixed length string
                            Me.RemoveLength(loParam)
                            loReturn.Parameters.Add(loParam)

                            '-- Increment the parameter counter
                            lnParamCnt += 1

                        Case WhereStatementTypeOptions.EndsWith
                            '-- Append the parametername
                            loQuery.Append("@PARAM")
                            loQuery.Append(lnParamCnt.ToString())

                            '-- Add the parameter to the return command
                            loParam = Me.CreateDbParameter(QueryInfo, loWhere.Fields(0))
                            loParam.ParameterName = "@PARAM" & lnParamCnt.ToString()
                            loParam.Value = "%" & CType(loWhere.Values(0), String)
                            loParam.Direction = ParameterDirection.Input
                            '-- Remove the length if this is a fixed length string
                            Me.RemoveLength(loParam)
                            loReturn.Parameters.Add(loParam)

                            '-- Increment the parameter counter
                            lnParamCnt += 1

                        Case WhereStatementTypeOptions.Equals, _
                                WhereStatementTypeOptions.GreaterThan, _
                                WhereStatementTypeOptions.GreaterThanOrEqual, _
                                WhereStatementTypeOptions.LessThan, _
                                WhereStatementTypeOptions.LessThanOrEqual, _
                                WhereStatementTypeOptions.Like

                            '-- Append the parametername
                            loQuery.Append("@PARAM")
                            loQuery.Append(lnParamCnt.ToString())

                            '-- Add the parameter to the return command
                            loParam = Me.CreateDbParameter(QueryInfo, loWhere.Fields(0))
                            loParam.ParameterName = "@PARAM" & lnParamCnt.ToString()
                            loParam.Value = loWhere.Values(0)
                            loParam.Direction = ParameterDirection.Input
                            loReturn.Parameters.Add(loParam)

                            '-- Increment the parameter counter
                            lnParamCnt += 1

                        Case WhereStatementTypeOptions.In
                            '-- Append the open parenthasis
                            loQuery.Append("(")

                            '-- Cycle through the parameters and add them
                            For Each loValue In loWhere.Values
                                '-- Append the parametername
                                loQuery.Append("@PARAM")
                                loQuery.Append(lnParamCnt.ToString())
                                loQuery.Append(", ")

                                '-- Add the parameter to the return command
                                loParam = Me.CreateDbParameter(QueryInfo, loWhere.Fields(0))
                                loParam.ParameterName = "@PARAM" & lnParamCnt.ToString()
                                loParam.Value = loValue
                                loParam.Direction = ParameterDirection.Input
                                loReturn.Parameters.Add(loParam)

                                '-- Increment the parameter counter
                                lnParamCnt += 1
                            Next loValue

                            '-- Remove the last 2 characters ", "
                            loQuery.Length -= 2

                            '-- Append the closing parenthasis
                            loQuery.Append(")")
                    End Select
                End If

                '-- Add the AND
                If QueryInfo.UseOrNotAnd Then
                    loQuery.Append(" OR ")
                Else
                    loQuery.Append(" AND ")
                End If
            Next loWhere

            '-- Remove the last 5 characters " AND "
            If QueryInfo.WhereStatements.Count > 0 Then
                If QueryInfo.UseOrNotAnd Then
                    loQuery.Length -= 4
                Else
                    loQuery.Length -= 5
                End If
            End If

            '-- Add the ORDER BY
            If QueryInfo.OrderByFields.Count > 0 Then
                loQuery.Append(" ORDER BY ")

                '-- Cycle through the order by fields and add them
                For Each lcField In QueryInfo.OrderByFields
                    loQuery.Append("`")
                    loQuery.Append(lcField)
                    loQuery.Append("`, ")
                Next

                '-- Remove the last 2 characters ", "
                loQuery.Length -= 2
            End If

            '-- Complete the command
            loQuery.AppendLine(";")

            '-- Set the command text
            loReturn.CommandText = ReplaceTokens(loQuery.ToString(), QueryInfo)
            loReturn.CommandType = CommandType.Text

            '-- Return the completed command
            Return loReturn
        End Function

        Private Function BuildUpdateCommand(ByVal QueryInfo As QueryInformation) As MySqlCommand
            '-- Determine whether this is a stored procedure or parameterized command
            If QueryInfo.UseStoredProcedure Then
                Return Me.BuildUpdateCommand_SP(QueryInfo)
            Else
                Return Me.BuildUpdateCommand_NonSP(QueryInfo)
            End If
        End Function

        ''' <summary>
        ''' Edhy Rijo 4/29/2010
        ''' MySql uses the character ` to wrap the UPDATE command fields.
        ''' The "Sql Server Mode=True" connection parameter could be used, but that may not be supported
        ''' by all .Net connectors
        ''' </summary>
        ''' <param name="QueryInfo"></param>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Private Function BuildUpdateCommand_NonSP(ByVal QueryInfo As QueryInformation) As MySqlCommand
            '-- Establish locals
            Dim loReturn As New MySqlCommand()
            Dim loParam As MySqlParameter
            Dim lcField As String
            Dim loQuery As New StringBuilder(1024)
            Dim loType As MySqlDbType

            '-- Begin building the command
            loQuery.Append("UPDATE `")
            loQuery.Append(QueryInfo.TableSchema)
            loQuery.Append("`.`")
            loQuery.Append(QueryInfo.TableName)
            loQuery.Append("` SET ")

            '-- Add the set statements
            For Each lcField In QueryInfo.Fields
                '-- Append the set to the query
                loQuery.Append("`")
                loQuery.Append(lcField)
                loQuery.Append("` = @")
                loQuery.Append(lcField)
                loQuery.Append(", ")

                '-- Add the parameter to the command
                loParam = Me.CreateDbParameter(QueryInfo, lcField)
                loParam.ParameterName = "@" & lcField
                loParam.SourceColumn = lcField
                loParam.Direction = ParameterDirection.Input
                loReturn.Parameters.Add(loParam)
            Next

            '-- Add the row version fields
            For Each lcField In QueryInfo.RowVersionFields
                '-- Determine the type of the row version information
                If QueryInfo.ConcurrencyType = ConcurrencyTypeOptions.OptimisticRowVersion Then
                    '-- Append the row version updator
                    loQuery.Append("`")
                    loQuery.Append(lcField)
                    loQuery.Append("` = (`")
                    loQuery.Append(lcField)
                    loQuery.Append("` + 1), ")
                End If
            Next

            '-- Remove the last 2 characters ", "
            loQuery.Length -= 2

            '-- Append the WHERE clause
            loQuery.Append(" WHERE ")

            '-- Append the primary key value(s)
            For Each lcField In QueryInfo.PrimaryKeyFields
                loQuery.Append("`")
                loQuery.Append(lcField)
                loQuery.Append("` = @")
                loQuery.Append(lcField)
                If Not QueryInfo.PrimaryKeyFieldIsAutoInc AndAlso QueryInfo.PrimaryKeyFieldIsUpdatable Then
                    loQuery.Append("_ORGPK")
                End If
                loQuery.Append(" AND ")

                '-- Add the parameter to the command
                loParam = Me.CreateDbParameter(QueryInfo, lcField)
                loParam.ParameterName = "@" & lcField
                If Not QueryInfo.PrimaryKeyFieldIsAutoInc AndAlso QueryInfo.PrimaryKeyFieldIsUpdatable Then
                    loParam.ParameterName &= "_ORGPK"
                End If
                loParam.SourceColumn = lcField
                loParam.Direction = ParameterDirection.Input
                loParam.SourceVersion = DataRowVersion.Original
                loReturn.Parameters.Add(loParam)
            Next

            '-- Remove the last " AND "
            loQuery.Length -= 5

            '-- Add the row version fields to the return value
            If QueryInfo.ConcurrencyType = ConcurrencyTypeOptions.OptimisticRowVersion OrElse _
                    QueryInfo.ConcurrencyType = ConcurrencyTypeOptions.OptimisticTimestamp Then

                For Each lcField In QueryInfo.RowVersionFields
                    '-- Only check the NULL on the parameter
                    loQuery.Append(" AND ((`")
                    loQuery.Append(lcField)
                    loQuery.Append("` = @")
                    loQuery.Append(lcField)
                    loQuery.Append(") OR (@")
                    loQuery.Append(lcField)
                    loQuery.Append(" IS NULL))")

                    '-- Append the parameter
                    loParam = Me.CreateDbParameter(QueryInfo, lcField)
                    loParam.ParameterName = "@" & lcField
                    loParam.SourceColumn = lcField
                    loParam.Direction = ParameterDirection.Input
                    loReturn.Parameters.Add(loParam)
                Next
            ElseIf QueryInfo.ConcurrencyType = ConcurrencyTypeOptions.OptimisticAllFields Then
                For Each lcField In QueryInfo.Fields
                    '-- Add the field to the query
                    '-- Make sure the field is not an image field
                    'Try
                    '    loType = CType(QueryInfo.FieldNativeDbTypes(lcField), MySqlDbType)
                    '    If loType = MySqlDbType.Image Then
                    '        Continue For
                    '    End If
                    'Catch ex As Exception
                    '    loType = MySqlDbType.VarChar
                    'End Try

                    loQuery.Append(" AND ((`")
                    loQuery.Append(lcField)
                    If (loType = MySqlDbType.Text) OrElse _
                        ((loType = MySqlDbType.VarChar) AndAlso (QueryInfo.FieldLengths(lcField) = -1)) OrElse _
                        ((QueryInfo.FieldLengths(lcField) = -1)) Then
                        loQuery.Append("` LIKE @")
                    Else
                        loQuery.Append("` = @")
                    End If
                    loQuery.Append(lcField)
                    loQuery.Append("_ORG) OR ((@")
                    loQuery.Append(lcField)
                    loQuery.Append("_ORG IS NULL) AND (`")
                    loQuery.Append(lcField)
                    loQuery.Append("` IS NULL)))")

                    '-- Add the original field value parameter
                    loParam = Me.CreateDbParameter(QueryInfo, lcField)
                    loParam.ParameterName = "@" & lcField & "_ORG"
                    loParam.SourceColumn = lcField
                    loParam.SourceVersion = DataRowVersion.Original
                    loParam.Direction = ParameterDirection.Input
                    loReturn.Parameters.Add(loParam)
                Next
            End If

            '-- Close off the UPDATE command
            loQuery.AppendLine(";")

            '-- Add the select command to retrieve the row version field
            If QueryInfo.RowVersionFields.Count > 0 Then
                '-- Build the beginning of the query
                loQuery.Append("SELECT ")

                '-- Add the row version fields
                For Each lcField In QueryInfo.RowVersionFields
                    loQuery.Append("`")
                    loQuery.Append(lcField)
                    loQuery.Append("`, ")
                Next

                '-- Remove the last 2 characters ", "
                loQuery.Length -= 2

                '-- Add the FROM
                loQuery.Append(" FROM `")
                loQuery.Append(QueryInfo.TableSchema)
                loQuery.Append("`.`")
                loQuery.Append(QueryInfo.TableName)
                loQuery.Append("` WHERE ")

                '-- Add the where and the primary key(s)
                For Each lcField In QueryInfo.PrimaryKeyFields
                    loQuery.Append("`")
                    loQuery.Append(lcField)
                    loQuery.Append("` = @")
                    loQuery.Append(lcField)
                    loQuery.Append(" AND ")
                Next

                '-- Remove the last " AND "
                loQuery.Length -= 5

                '-- Close off the command
                loQuery.AppendLine(";")
            End If

            '-- Set the command text
            loReturn.CommandText = ReplaceTokens(loQuery.ToString(), QueryInfo)
            loReturn.CommandType = CommandType.Text

            '-- Return the command
            Return loReturn
        End Function

        Private Function BuildUpdateCommand_SP(ByVal QueryInfo As QueryInformation) As MySqlCommand
            '-- Establish locals
            Dim loReturn As New MySqlCommand()
            Dim loParam As MySqlParameter
            Dim lcField As String

            '-- Begin building the command by setting the text
            loReturn.CommandText = QueryInfo.StoredProcedureName
            loReturn.CommandType = CommandType.StoredProcedure

            '-- Add the parameters to the command
            For Each lcField In QueryInfo.Fields
                '-- Add the field values
                loParam = Me.CreateDbParameter(QueryInfo, lcField)
                loParam.ParameterName = "@" & lcField
                loParam.SourceColumn = lcField
                loParam.Direction = ParameterDirection.Input
                loReturn.Parameters.Add(loParam)
            Next

            '-- Add the primary key field
            For Each lcField In QueryInfo.PrimaryKeyFields
                loParam = Me.CreateDbParameter(QueryInfo, lcField)
                loParam.ParameterName = "@" & lcField
                If Not QueryInfo.PrimaryKeyFieldIsAutoInc AndAlso QueryInfo.PrimaryKeyFieldIsUpdatable Then
                    loParam.ParameterName &= "_ORGPK"
                    loParam.SourceVersion = DataRowVersion.Original
                End If
                loParam.SourceColumn = lcField
                loParam.Direction = ParameterDirection.Input
                loReturn.Parameters.Add(loParam)
            Next

            '-- Add the row version fields
            If QueryInfo.ConcurrencyType = ConcurrencyTypeOptions.OptimisticRowVersion OrElse _
                QueryInfo.ConcurrencyType = ConcurrencyTypeOptions.OptimisticTimestamp Then

                '-- Cycle through the fields
                For Each lcField In QueryInfo.RowVersionFields
                    '-- Add the field parameter
                    loParam = Me.CreateDbParameter(QueryInfo, lcField)
                    loParam.ParameterName = "@" & lcField
                    loParam.SourceColumn = lcField
                    loParam.Direction = ParameterDirection.InputOutput
                    loReturn.Parameters.Add(loParam)
                Next
            ElseIf QueryInfo.ConcurrencyType = ConcurrencyTypeOptions.OptimisticAllFields Then
                '-- Cycle through the fields and add the _ORG parameters
                For Each lcField In QueryInfo.Fields
                    '-- Add the field parameter
                    loParam = Me.CreateDbParameter(QueryInfo, lcField)
                    loParam.ParameterName = "@" & lcField & "_ORG"
                    loParam.SourceColumn = lcField
                    loParam.Direction = ParameterDirection.Input
                    loParam.SourceVersion = DataRowVersion.Original
                    loReturn.Parameters.Add(loParam)
                Next
            End If

            '-- Return the command
            Return loReturn
        End Function

        Private Function CreateDbParameter(ByVal QueryInfo As QueryInformation, ByVal FieldName As String) As MySqlParameter
            '-- Establish locals
            Dim loReturn As New MySqlParameter()

            '-- Set data type on the parameter
            Try
                '-- First set the native db type
                loReturn.MySqlDbType = CType(QueryInfo.FieldNativeDbTypes(FieldName), MySqlDbType)

                '-- Check to make sure the dbtype matches up
                If loReturn.DbType <> QueryInfo.FieldDbTypes(FieldName) Then

                    '-- The dbtype is off, so force it to match up
                    loReturn.DbType = QueryInfo.FieldDbTypes(FieldName)
                End If

                '-- Set the length on the parameter
                If loReturn.MySqlDbType = MySqlDbType.Binary OrElse _
                    loReturn.MySqlDbType = MySqlDbType.VarBinary OrElse _
                    loReturn.MySqlDbType = MySqlDbType.VarChar Then

                    '-- TLT - 01/17/2008 - Made a change to add two to the length of the size so that small character fields
                    '   such as Char(2) would support the values entered for a BrowseDialog.  If a contains was used and the letters
                    '   'it' were set, the query value would be '%i' instead of '%it%'  By adjusting this value, it appears to have fixed the problems.

                    '-- Set the size on the parameter
                    If QueryInfo.FieldLengths(FieldName) > 0 Then
                        loReturn.Size = QueryInfo.FieldLengths(FieldName) + 2
                    Else
                        loReturn.Size = QueryInfo.FieldLengths(FieldName)
                    End If
                    '-- END OF TLT CHANGE
                End If
            Catch ex As Exception
                '-- Set the dbtype only
                loReturn.DbType = QueryInfo.FieldDbTypes(FieldName)
            End Try

            '-- Return the created parameter
            Return loReturn
        End Function

        Private Sub RemoveLength(ByVal Parameter As MySqlParameter)
            '-- Test the parameter type
            If Parameter.DbType = DbType.AnsiStringFixedLength OrElse _
                Parameter.DbType = DbType.StringFixedLength Then

                '-- Set the size on the parameter
                Parameter.Size = -1
            End If
        End Sub

        ''' <summary>
        ''' Replaces any tokens place holders with the desired values
        ''' </summary>
        Private Function ReplaceTokens(ByVal queryString As String, ByVal queryInfo As QueryInformation) As String
            '-- Establish Locals
            Dim ret As String = queryString

            '-- Replace all tokens
            For Each keyPair As KeyValuePair(Of String, String) In queryInfo.Tokens
                ret = ret.Replace("`" & keyPair.Key & "`", keyPair.Value)
            Next

            '-- Return results
            Return ret
        End Function

#End Region

#Region " Protected Methods "

        ''' <summary>
        ''' Updates the data source with the information from the specified row and executes the command.
        ''' </summary>
        ''' <param name="QueryInfo"></param>
        ''' <param name="RowToUpdate"></param>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Protected Overrides Function UpdateRow(ByVal QueryInfo As QueryInformation, ByVal RowToUpdate As System.Data.DataRow, ByVal ConcurrencyHandler As ConcurrencyExceptionHandler, ByVal RowErrorHandler As AddRowErrorHandler, ByVal RecreateCommand As Boolean) As Integer
            '-- Establish locals
            Dim loCommand As MySqlCommand = Nothing
            Dim lnReturn As Integer
            Dim loParam As MySqlParameter
            Dim lcField As String
            Dim loReader As IDataReader
            Dim loReSaveQueryInfo As QueryInformation
            Dim llIsCommandCached As Boolean

            '-- Determine the type of command
            Select Case RowToUpdate.RowState
                Case DataRowState.Deleted

                    '-- Lock the dictionary
                    SyncLock Me._CachedDeleteCommands

                        '-- Get the command to execute
                        llIsCommandCached = Me._CachedDeleteCommands.ContainsKey(RowToUpdate.Table)
                        If (Not llIsCommandCached) OrElse (RecreateCommand) Then
                            '-- Create a new command
                            loCommand = Me.BuildDeleteCommand(QueryInfo)

                            '-- Set the connection 
                            If Me.CanConnectDirectly Then
                                If QueryInfo.IsTransactional Then
                                    loCommand.Transaction = CType(Me.Transactions(QueryInfo.TransactionKey), MySqlTransaction)
                                    loCommand.Connection = CType(Me.Transactions(QueryInfo.TransactionKey).Connection, MySqlConnection)
                                Else
                                    loCommand.Connection = CType(Me.CreateBlankDbConnection(), MySqlConnection)
                                    loCommand.Connection.Open()
                                End If
                            End If
                            If Not llIsCommandCached Then
                                Me._CachedDeleteCommands.Add(RowToUpdate.Table, loCommand)
                            End If
                        Else
                            loCommand = Me._CachedDeleteCommands(RowToUpdate.Table)
                        End If

                        '-- Get a clone of the command to avoid threadlocking
                        loCommand = loCommand.Clone()
                    End SyncLock

                    '-- Set the values on the command parameters
                    For Each loParam In loCommand.Parameters
                        '-- Get the value and set it
                        loParam.Value = RowToUpdate.Item(loParam.SourceColumn, DataRowVersion.Original)
                    Next

                    '-- Execute the command 
                    lnReturn = Me.InternalExecuteNonQuery(loCommand, QueryInfo.IsTransactional, QueryInfo.TransactionKey)

                    '-- If there was a concurrency exception, then handle the exception and re-save the row
                    If (lnReturn < 1) Then

                        Select Case ConcurrencyHandler.Invoke(RowToUpdate)
                            Case ConcurrencyExceptionHandlingOptions.ForceSave
                                '-- Re-save the row after the collisions have been handled
                                '-- Remove any error on the row
                                lnReturn = Me.UpdateRow(QueryInfo, RowToUpdate, ConcurrencyHandler, RowErrorHandler)

                            Case ConcurrencyExceptionHandlingOptions.IgnoreSetError
                                '-- Just return 1, but set the row in error so that the changes will
                                '   not be accepted
                                lnReturn = 1
                                RowErrorHandler(RowToUpdate, "Row not updated.")
                                'RowToUpdate.RowError = "Row not udpated."

                            Case ConcurrencyExceptionHandlingOptions.IgnoreNoError
                                '-- Just return 1 so that the changes will be accepted on the row
                                lnReturn = 1

                        End Select
                    End If

                Case DataRowState.Added
                    '-- Lock the dictionary
                    SyncLock Me._CachedInsertCommands

                        '-- Get the command to execute
                        If Not Me._CachedInsertCommands.ContainsKey(RowToUpdate.Table) Then
                            '-- Create a new command
                            loCommand = Me.BuildInsertCommand(QueryInfo)

                            If Me.CanConnectDirectly Then
                                '-- See if this is transactional
                                If QueryInfo.IsTransactional Then
                                    loCommand.Transaction = CType(Me.Transactions(QueryInfo.TransactionKey), MySqlTransaction)
                                    loCommand.Connection = CType(Me.Transactions(QueryInfo.TransactionKey).Connection, MySqlConnection)
                                Else
                                    loCommand.Connection = CType(Me.CreateBlankDbConnection(), MySqlConnection)
                                    loCommand.Connection.Open()
                                End If
                            End If

                            '-- Save the item off in a dictionary
                            Me._CachedInsertCommands.Add(RowToUpdate.Table, loCommand)
                        Else
                            loCommand = Me._CachedInsertCommands(RowToUpdate.Table)
                        End If

                        '-- Get a clone of the command to avoid thread locking
                        loCommand = loCommand.Clone()

                    End SyncLock

                    '-- Set the values on the command parameters
                    For Each loParam In loCommand.Parameters
                        '-- Get the value and set it
                        loParam.Value = RowToUpdate.Item(loParam.SourceColumn, loParam.SourceVersion)
                    Next

                    '-- We need to execute this and gather back the value from the new pk
                    If QueryInfo.UseStoredProcedure Then

                        '-- Execute the command non query
                        lnReturn = Me.InternalExecuteNonQuery(loCommand, QueryInfo.IsTransactional, QueryInfo.TransactionKey)

                        '-- Get the new primary key value
                        If QueryInfo.PrimaryKeyFieldIsAutoInc Then
                            Me.UpdateFieldOnRow(RowToUpdate, QueryInfo.PrimaryKeyFields(0), Me.GetValueFromCommandParameter(loCommand, QueryInfo.PrimaryKeyFields(0)))
                        End If

                        '-- Get the new row version values
                        For Each lcField In QueryInfo.RowVersionFields
                            '-- Copy the value to the row
                            Me.UpdateFieldOnRow(RowToUpdate, lcField, Me.GetValueFromCommandParameter(loCommand, lcField))
                        Next

                    Else
                        '-- Execute the first command reader
                        'SyncLock _ReaderLock
                        loReader = Me.InternalExecuteReader(loCommand, QueryInfo.IsTransactional, QueryInfo.TransactionKey)

                        '-- Get the number of rows affected
                        lnReturn = loReader.RecordsAffected

                        '-- Execute the second command to get the info
                        If lnReturn > 0 Then
                            '-- Read the record from the reader
                            loReader.Read()

                            '-- Get the new primary key value
                            If QueryInfo.PrimaryKeyFieldIsAutoInc Then
                                Me.UpdateFieldOnRow(RowToUpdate, QueryInfo.PrimaryKeyFields(0), loReader.Item(QueryInfo.PrimaryKeyFields(0)))
                            End If

                            '-- Get the new row version values
                            For Each lcField In QueryInfo.RowVersionFields
                                '-- Copy the value to the row
                                Me.UpdateFieldOnRow(RowToUpdate, lcField, loReader.Item(lcField))
                            Next
                        End If

                        '-- Close the reader
                        loReader.Close()
                        'End SyncLock
                    End If

                Case DataRowState.Modified
                    '-- Lock the dictionary
                    SyncLock Me._CachedUpdateCommands

                        '-- Get the command to execute
                        llIsCommandCached = Me._CachedUpdateCommands.ContainsKey(RowToUpdate.Table)
                        If (Not llIsCommandCached) OrElse (RecreateCommand) Then
                            '-- Create a new command
                            loCommand = Me.BuildUpdateCommand(QueryInfo)

                            '-- Set the connection
                            If Me.CanConnectDirectly Then
                                If QueryInfo.IsTransactional Then
                                    loCommand.Transaction = CType(Me.Transactions(QueryInfo.TransactionKey), MySqlTransaction)
                                    loCommand.Connection = CType(Me.Transactions(QueryInfo.TransactionKey).Connection, MySqlConnection)
                                Else
                                    loCommand.Connection = CType(Me.CreateBlankDbConnection(), MySqlConnection)
                                    loCommand.Connection.Open()
                                End If
                            End If

                            '-- Add the item to the cache dictionary
                            If Not llIsCommandCached Then
                                Me._CachedUpdateCommands.Add(RowToUpdate.Table, loCommand)
                            End If
                        Else
                            loCommand = Me._CachedUpdateCommands(RowToUpdate.Table)
                        End If

                        '-- Get a clone to avoid thread locking
                        loCommand = loCommand.Clone()

                    End SyncLock

                    '-- Set the values on the command parameters
                    For Each loParam In loCommand.Parameters
                        '-- Get the value and set it
                        loParam.Value = RowToUpdate.Item(loParam.SourceColumn, loParam.SourceVersion)
                    Next

                    If QueryInfo.UseStoredProcedure Then
                        '-- We need to execute the command non query
                        lnReturn = Me.InternalExecuteNonQuery(loCommand, QueryInfo.IsTransactional, QueryInfo.TransactionKey)

                        '-- See if the rows were affected
                        If lnReturn > 0 Then
                            '-- Retrieve the new timestamp/rowversion values
                            For Each lcField In QueryInfo.RowVersionFields
                                '-- Save the new value
                                Me.UpdateFieldOnRow(RowToUpdate, lcField, Me.GetValueFromCommandParameter(loCommand, lcField))
                            Next
                        End If
                    Else
                        'SyncLock _ReaderLock
                        '-- We need to execute the first command reader because it's a batch
                        loReader = Me.InternalExecuteReader(loCommand, QueryInfo.IsTransactional, QueryInfo.TransactionKey)

                        '-- Get the rows affected
                        lnReturn = loReader.RecordsAffected

                        '-- Get the new timestamp/rowversion values
                        If lnReturn > 0 Then
                            loReader.Read()

                            '-- Get 'em
                            For Each lcField In QueryInfo.RowVersionFields
                                '-- Save the new value
                                Me.UpdateFieldOnRow(RowToUpdate, lcField, loReader.Item(lcField))
                            Next
                        End If

                        '-- Close off the reader
                        loReader.Close()
                        'End SyncLock
                    End If

                    '-- If there was a concurrency exception, then handle the exception and re-save the row
                    If (lnReturn < 1) Then

                        Select Case ConcurrencyHandler.Invoke(RowToUpdate)
                            Case ConcurrencyExceptionHandlingOptions.ForceSave
                                '-- Re-save the row after the collisions have been handled
                                '-- Turn off the concurrency if it was set to all fields
                                If QueryInfo.ConcurrencyType = ConcurrencyTypeOptions.OptimisticAllFields Then
                                    loReSaveQueryInfo = QueryInfo.Clone()
                                    loReSaveQueryInfo.ConcurrencyType = ConcurrencyTypeOptions.Off
                                    lnReturn = Me.UpdateRow(loReSaveQueryInfo, RowToUpdate, ConcurrencyHandler, RowErrorHandler, True)
                                Else
                                    lnReturn = Me.UpdateRow(QueryInfo, RowToUpdate, ConcurrencyHandler, RowErrorHandler)
                                End If

                            Case ConcurrencyExceptionHandlingOptions.IgnoreSetError
                                '-- Just return 1, but set the row in error so that the changes will
                                '   not be accepted
                                lnReturn = 1
                                RowErrorHandler(RowToUpdate, "Row not updated.")
                                'RowToUpdate.RowError = "Row not udpated."

                            Case ConcurrencyExceptionHandlingOptions.IgnoreNoError
                                '-- Just return 1 so that the changes will be accepted on the row
                                lnReturn = 1

                        End Select

                    End If
            End Select

            '-- Dispose of the cloned command
            loCommand.Connection = Nothing
            loCommand.Dispose()

            '-- Return the rows affected
            Return lnReturn
        End Function

#End Region

#Region " Public Methods "

        ''' <summary>
        ''' Returns a new MySqlCommand object for use with this DataSource.
        ''' </summary>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Public Overrides Function CreateBlankDbCommand() As System.Data.Common.DbCommand
            '-- Return a new MySqlCommand
            Return New MySqlCommand()
        End Function

        ''' <summary>
        ''' Returns a new MySqlConnection that can be used to connect to this DataSource.
        ''' </summary>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Public Overrides Function CreateBlankDbConnection() As System.Data.Common.DbConnection
            '-- Return a new MySqlConnection
            Return New MySqlConnection(Me.ConnectionString)
        End Function

        ''' <summary>
        ''' Returns a new SqlDataAdapter that can be used to connect to this DataSource.
        ''' </summary>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Public Overrides Function CreateBlankDbDataAdapter() As System.Data.Common.DbDataAdapter
            '-- Return a new SqlDataAdapter
            'Return New SqlDataAdapter()
            Return New MySqlDataAdapter()
        End Function

        ''' <summary>
        ''' Returns a new MySqlParameter that can be used with this DataSource.
        ''' </summary>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Public Overrides Function CreateBlankDbParameter() As System.Data.Common.DbParameter
            '-- Return a new MySqlParameter
            'Return New SqlParameter()
            Return New MySqlParameter()
        End Function

        ''' <summary>
        ''' Returns a new QueryCommand that matches the given QueryInformation.
        ''' </summary>
        ''' <param name="QueryInfo"></param>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Public Overrides Function CreateDbCommand(ByVal QueryInfo As QueryInformation) As DbCommand
            '-- Establish locals
            Dim loReturn As MySqlCommand = Nothing

            '-- Return the results based upon the given queryinfo
            Select Case QueryInfo.QueryType
                Case QueryTypeOptions.DELETE
                    loReturn = Me.BuildDeleteCommand(QueryInfo)

                Case QueryTypeOptions.INSERT
                    loReturn = Me.BuildInsertCommand(QueryInfo)

                Case QueryTypeOptions.SELECT, QueryTypeOptions.SELECT_TOP
                    loReturn = Me.BuildSelectCommand(QueryInfo)

                Case QueryTypeOptions.UPDATE
                    loReturn = Me.BuildUpdateCommand(QueryInfo)
            End Select

            '-- Return the result
            Return loReturn
        End Function

        ''' <summary>
        ''' Clears the command cache for the given data table.
        ''' </summary>
        ''' <param name="UpdatingTable"></param>
        ''' <remarks></remarks>
        Public Overrides Sub EndUpdateCommandCaching(ByVal UpdatingTable As System.Data.DataTable, ByVal Transactional As Boolean)
            '-- Establish locals
            Dim loCommand As MySqlCommand

            '-- Clear out the dictionary entries
            If Me._CachedDeleteCommands.ContainsKey(UpdatingTable) Then
                '-- Close the connection on the command
                loCommand = Me._CachedDeleteCommands(UpdatingTable)

                '-- Make sure that we were not on a transaction
                If (Not Transactional) AndAlso (loCommand.Connection IsNot Nothing) Then
                    loCommand.Connection.Close()
                    loCommand.Connection.Dispose()
                End If
                loCommand.Dispose()

                '-- Remove it from the dictionary
                Me._CachedDeleteCommands.Remove(UpdatingTable)
            End If

            If Me._CachedInsertCommands.ContainsKey(UpdatingTable) Then
                '-- Close the connection on the command
                loCommand = Me._CachedInsertCommands(UpdatingTable)

                '-- Make sure that we were not on a transaction
                If (Not Transactional) AndAlso (loCommand.Connection IsNot Nothing) Then
                    loCommand.Connection.Close()
                    loCommand.Connection.Dispose()
                End If
                loCommand.Dispose()

                '-- Remove it from the dictionary
                Me._CachedInsertCommands.Remove(UpdatingTable)
            End If

            If Me._CachedUpdateCommands.ContainsKey(UpdatingTable) Then
                '-- Close the connection on the command
                loCommand = Me._CachedUpdateCommands(UpdatingTable)

                '-- Make sure that we were not on a transaction
                If (Not Transactional) AndAlso (loCommand.Connection IsNot Nothing) Then
                    loCommand.Connection.Close()
                    loCommand.Connection.Dispose()
                End If
                loCommand.Dispose()

                '-- Remove it from the dictionary
                Me._CachedUpdateCommands.Remove(UpdatingTable)
            End If
        End Sub

        ''' <summary>
        ''' Returns the Database name for this DataSource.
        ''' </summary>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Public Overrides Function GetDatabaseName() As String
            '-- Return the database name
            'Return New SqlConnectionStringBuilder(Me.ConnectionString).InitialCatalog

            Return New MySqlConnectionStringBuilder(Me.ConnectionString).Database
        End Function

        ' ''' <summary>
        ' ''' Executes the given command on the data source and returns the results.
        ' ''' </summary>
        ' ''' <param name="Command"></param>
        ' ''' <param name="CallBack"></param>
        ' ''' <returns></returns>
        ' ''' <remarks></remarks>
        'Public Overloads Overrides Function GetDataTable(ByVal Command As System.Data.Common.DbCommand, ByVal CallBack As OnChangeEventHandler) As System.Data.DataTable
        '    ' Public Overloads Overrides Function GetDataTable(ByVal Command As System.Data.Common.DbCommand, ByVal CallBack As OnChangeEventHandler) As System.Data.DataTable
        '    '-- Establish locals
        '    'Dim loDep As SqlDependency

        '    ''-- Add the handler to the dependency
        '    'If CallBack IsNot Nothing Then
        '    '    loDep = New SqlDependency()

        '    '    AddHandler loDep.OnChange, CallBack
        '    '    loDep.AddCommandDependency(CType(Command, SqlCommand))
        '    '    Try
        '    '        'SqlDependency.Start(Command.Connection.ConnectionString)
        '    '        SqlDependency.Start(CreateBlankDbConnection().ConnectionString)
        '    '    Catch ex As Exception
        '    '        MsgBox(ex.Message)
        '    '    End Try
        '    'End If

        '    '-- Execute the command
        '    Return MyBase.GetDataTable(Command, CallBack)
        'End Function

        ' Dim x As System.Data.SqlClient.SqlDependency()
        ' Edhy to be checked.
        'Dim xx As MySql.Data.MySqlClient.mysqldependency



        ' ''' <summary>
        ' ''' Gets the native MySqlDbType of the given parameter.
        ' ''' </summary>
        ' ''' <param name="Parameter"></param>
        ' ''' <returns></returns>
        ' ''' <remarks></remarks>
        'Public Overrides Function GetNativeDbType(ByVal Parameter As MySqlParameter) As Integer
        '    'Return CType(CType(Parameter, SqlParameter).SqlDbType, Integer)
        '    Return CType(CType(Parameter, MySqlParameter).MySqlDbType, Integer)
        'End Function

        Public Overrides Function GetNativeDbType(ByVal Parameter As System.Data.Common.DbParameter) As Integer
            Return CType(CType(Parameter, MySqlParameter).MySqlDbType, Integer)
        End Function


        ''' <summary>
        ''' Gets the native OleDbType for the specified field.
        ''' </summary>
        ''' <param name="QueryInfo"></param>
        ''' <param name="FieldName"></param>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Private Overloads Function GetNativeDbType(ByVal QueryInfo As QueryInformation, ByVal FieldName As String) As MySqlDbType
            ''  Private Overloads Function GetNativeDbType(ByVal QueryInfo As QueryInformation, ByVal FieldName As String) As OleDbType
            If Not QueryInfo.ShouldNativeDbTypesBeConverted Then
                '-- The type does not need to be converted
                Return CType(QueryInfo.FieldNativeDbTypes(FieldName), System.Data.DbType)
            Else
                '-- The type is SqlDbTYpe and needs to be converted
                Select Case CType(QueryInfo.FieldNativeDbTypes(FieldName), MySqlDbType)
                    Case SqlDbType.BigInt
                        Return MySqlDbType.Int64
                    Case SqlDbType.Binary
                        Return MySqlDbType.Binary
                    Case SqlDbType.Bit
                        Return MySqlDbType.Binary
                    Case SqlDbType.Char
                        Return MySqlDbType.VarChar
                    Case SqlDbType.DateTime
                        Return MySqlDbType.DateTime
                    Case SqlDbType.Decimal
                        Return MySqlDbType.Decimal
                    Case SqlDbType.Float
                        Return MySqlDbType.Double
                        'Return MySqlDbType.Numeric
                    Case SqlDbType.Image
                        Return MySqlDbType.LongBlob
                    Case SqlDbType.Int
                        Return MySqlDbType.Int32
                    Case SqlDbType.Money
                        Return MySqlDbType.Decimal
                    Case SqlDbType.NChar
                        Return MySqlDbType.VarChar
                    Case SqlDbType.NText
                        Return MySqlDbType.LongText
                    Case SqlDbType.NVarChar
                        Return MySqlDbType.VarChar
                        'Case SqlDbType.Real
                        '    Return MySqlDbType.Singl
                    Case SqlDbType.SmallDateTime
                        Return MySqlDbType.Date
                    Case SqlDbType.SmallInt
                        Return MySqlDbType.Int16
                    Case SqlDbType.SmallMoney
                        Return MySqlDbType.Decimal
                    Case SqlDbType.Text
                        Return MySqlDbType.LongText
                    Case SqlDbType.Timestamp
                        Return MySqlDbType.Timestamp
                    Case SqlDbType.TinyInt
                        Return MySqlDbType.Int16
                        'Case SqlDbType.Udt
                        '    Return MySqlDbType.Variant
                    Case SqlDbType.UniqueIdentifier
                        Return MySqlDbType.Guid
                    Case SqlDbType.VarBinary
                        Return MySqlDbType.VarBinary
                    Case SqlDbType.VarChar
                        Return MySqlDbType.VarChar
                        'Case SqlDbType.Variant
                        '    Return MySqlDbType.Variant
                        'Case SqlDbType.Xml
                        '    Return MySqlDbType.Variant
                End Select
            End If
        End Function
        ''' <summary>
        ''' Sets the MySqlDbType on the given parameter.
        ''' </summary>
        ''' <param name="Parameter"></param>
        ''' <param name="NativeDbType"></param>
        ''' <remarks></remarks>
        Public Overrides Sub SetParameterNativeType(ByVal Parameter As DbParameter, ByVal NativeDbType As Integer)
            CType(Parameter, MySqlParameter).MySqlDbType = CType(NativeDbType, MySqlDbType)
        End Sub

#End Region


    End Class
End Namespace

